from importlib.metadata import version

from .decorators import job, worker
from .job import Cancel, Job, Record, Snooze
from .oban import Oban

try:
    import oban_pro  # noqa: F401  # ty: ignore[unresolved-import]
except ImportError:
    pass

__all__ = [
    "Cancel",
    "Job",
    "Oban",
    "Record",
    "Snooze",
    "job",
    "worker",
]

__version__ = version("oban")
